/* main.c -- Dune Webserver
 * Copyright (C) 1998 Baris Mestanogullari <mestanba@martin.luther.edu>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.  If not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <sys/socket.h>
#include <sys/types.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <errno.h>
#include <alloca.h>
#include <string.h>
#include <netdb.h>
#include <unistd.h>
#include <grp.h>

#include "main.h"
#include "config.h"

void sh() {
  free(buffer);
  free(received);
  close(sfd); 
  close(cfd); 
  close((int)urlp);
  fclose(logfile);
  fclose(errfile);
}

void term() {
#ifdef LOG_ERRORS
  fprintf(errfile, "%s: Got SIGTERM, shutting down.\n", currentTime());
#endif
  sh();
}

void segv() {
#ifdef LOG_ERRORS
  fprintf(errfile, "%s: Got SIGSEGV, dumping core.\n", currentTime());
#endif
  sh();
}

void setguid(char *name) {
  if(geteuid()==0) {
    if(UID!=0) {
      if(setgid(GID)==-1) {
        fprintf(stderr, "%s: Couldn't set GID, exiting.\n", name);
        exit(-1);
      }
      if(setuid(UID)==-1) {
        fprintf(stderr, "%s: Couldn't set UID, exiting.\n", name);
        exit(-1);
      }
    }
    else {
      fprintf(stderr, "%s: You cannot run Dune as root. Please change \
the UID and GID in config.h or start it up as a different user.\n", name);
      exit(-1);
    }
  }
}

void main(int argc, char**argv) {
  struct sockaddr_in ls,cs;  /* Sockets: ls=listen  cs=connected */
  struct hostent *he;  
  long int port=PORT;
  
  buffer=(char *)malloc(4096);
  received=(char *)malloc(512);

  /* Setup the signal handlers */
  signal(SIGTERM,&term);        /* Call term() when we receive SIGTERM */
  signal(SIGSEGV,&segv);        /* Call segv() when we receive SIGSEGV */
  signal(SIGCHLD,SIG_IGN);	/* Zombies should leave us alone now   */

#ifdef LOG_CONNECTIONS
  if((logfile=fopen(LOGFILE, "a"))==NULL) {
    fprintf(stderr, "%s: Can't open %s for writing.\n", argv[0], LOGFILE);
    exit(-1);
  }
#endif
#ifdef LOG_ERRORS
  if((errfile=fopen(ERROR_LOGFILE, "a"))==NULL) {
    fprintf(stderr, "%s: Can't open %s for writing.\n",
            argv[0], ERROR_LOGFILE);
    exit(-1);
  }
#endif

  if((sfd=socket(AF_INET,SOCK_STREAM,0))<0) {
    fprintf(stderr, "%s: Can't allocate socket\n",argv[0]);
    exit(-1);
  }
  
  ls.sin_family=AF_INET;
  ls.sin_addr.s_addr=INADDR_ANY;

  if(argc!=1 && argc!=3) {
    fprintf(stderr, "Usage: %s [-port port-number]\n", argv[0]);
    exit(-1);
  }
  if(argc==3) {
    if(!strcmp(argv[1], "-port")) {
      port=strtol(argv[2], (char **)NULL, 10);
      if(port==LONG_MIN || port==LONG_MAX || port==0) {
        fprintf(stderr, "Usage: %s [-port port-number]\n", argv[0]);
        exit(-1);
      }
    }
    else {
      fprintf(stderr, "Usage: %s [-port port-number]\n", argv[0]);
      exit(-1);
    }
  }

  ls.sin_port=htons(port);
  
  if(bind(sfd,(struct sockaddr *)&ls,sizeof(ls))<0) {
    if(port<1024 && geteuid()!=0)
      fprintf(stderr, "%s: Ports below 1024 can only be bound by root.\n",argv[0]);
    else fprintf(stderr, "%s: Port already in use.\n", argv[0]);
    exit(-1);
  }
  if(listen(sfd,5)<0) {
    fprintf(stderr, "%s: Can't listen for connections.\n", argv[0]);
    exit(-1);
  }

  setguid(argv[0]);
  printf("Dune %s started on port %li.\n", VERSION, port);
  if (fork()!=0) exit(0);
  setsid();

  while(1) {
    int al;
    al=sizeof(cs);
    if((cfd=accept(sfd,(struct sockaddr *)&cs,&al))<0) {
/*
 * From asm/errno.h:
 * #define     ECONNRESET      104     /\* Connection reset by peer *\/
 */
      if(errno!=104)
        exit(-1);
    }
    
    if (fork()==0) {
      fd_set temp;
      struct timeval tv;
      int rv;
      char remoteAddress[128];
      close(sfd);	/* Close the listening socket. */
      bzero(received,256);
      received[0]='\0';

      FD_ZERO(&temp);
      FD_SET(cfd, &temp);
      tv.tv_sec = 30;
      tv.tv_usec = 0;

      rv=select(cfd+1, &temp, NULL, NULL, &tv);

      if (!rv) exit(-1);
      buffer[0]='\0';
      if(read(cfd,buffer,1)==0) exit(0);
      while(1) {
	if(buffer[0]==EOF) exit(0);
	if(buffer[0]=='\n') break;
	strcat(received, buffer);
	buffer[0]='\0';
	read(cfd,buffer,1);
      }

      FD_ZERO(&temp);
      FD_SET(cfd, &temp);
      tv.tv_sec = 30;
      tv.tv_usec = 0;

      rv=select(cfd+1, &temp, NULL, NULL, &tv);

      if (!rv) exit(0);
      buffer[0]='\0';
      if(read(cfd,buffer,1)==0) exit(0);
      while(1) {
	if(buffer[0]==EOF) exit(0);
	if(buffer[0]=='\n') break;
	buffer[0]='\0';
	read(cfd,buffer,1);
      }

      /* Get the remote address */
      he=gethostbyaddr((char *)&cs.sin_addr,
		       sizeof(cs.sin_addr),cs.sin_family);
      if(he) strcpy(remoteAddress,he->h_name);
      else strcpy(remoteAddress,(char *)inet_ntoa(cs.sin_addr));

      processRequest(received, remoteAddress);
      write(cfd,NULL,(int)NULL);
      close(cfd);
      exit(0);
    }
    else close(cfd);
  }
}
